@@ -7,13 +7,13 @@ from __future__ import annotations
 from dataclasses import dataclass
 from itertools import chain
 from math import ceil, exp
-from typing import Any, Literal, Sequence, cast, overload
+from typing import Any, Literal, Sequence, cast, overload, TypeVar
 
 from vstools import (
     MISSING, ColorRange, ConstantFormatVideoNode, CustomIntEnum, CustomOverflowError, CustomStrEnum, CustomValueError,
     FieldBased, FieldBasedT, FuncExceptT, GenericVSFunction, InvalidColorFamilyError, MissingT, VSFunction,
     check_ref_clip, check_variable, core, depth, disallow_variable_format, disallow_variable_resolution, fallback,
-    kwargs_fallback, normalize_planes, normalize_seq, vs
+    kwargs_fallback, normalize_planes, normalize_seq, vs, scale_value, clamp
 )
 
 from .prefilters import PelType, Prefilter, prefilter_to_full_range
@@ -259,6 +259,13 @@ class SADMode(CustomIntEnum):
 
         return self >= SADMode.SATD
 
+    @property
+    def same_recalc(self: SelfSADMode) -> tuple[SelfSADMode, SelfSADMode]:
+        return (self, self)
+
+
+SelfSADMode = TypeVar('SelfSADMode', bound=SADMode)
+
 
 class MotionMode:
     """
@@ -317,7 +324,7 @@ class MotionMode:
         Note that vector length is smaller at lower level.
         """
 
-        global_motion: bool
+        pglobal: bool
         """
         Relative penalty (scaled to 8 bit) to SAD cost for global predictor vector.\n
         Coherence is not used for global vector.
@@ -335,7 +342,7 @@ class MotionMode:
     class _CustomConfig:
         def __call__(
             self, coherence: int | None = None, sad_limit: int | None = None,
-            pnew: int | None = None, plevel: int | None = None, global_motion: bool | None = None,
+            pnew: int | None = None, plevel: int | None = None, pglobal: bool | None = None,
             truemotion: bool = True
         ) -> MotionMode.Config:
             """
@@ -353,7 +360,7 @@ class MotionMode:
                 fallback(sad_limit, ref.sad_limit),
                 fallback(pnew, ref.pnew),
                 fallback(plevel, ref.plevel),
-                fallback(global_motion, ref.global_motion)
+                fallback(pglobal, ref.pglobal)
             )
 
     MANUAL = _CustomConfig()
@@ -380,13 +387,16 @@ class SearchModeBase:
         mode: SearchMode
         """SearchMode that decides which analysis mode to use for search of motion vectors."""
 
-        param: int
+        recalc_mode: SearchMode
+        """SearchMode that decides which analysis mode to use for recalculation of motion vectors."""
+
+        param: int | None
         """Parameter used by the search mode in analysis."""
 
-        param_recalc: int
+        param_recalc: int | None
         """Parameter used by the search mode in recalculation."""
 
-        pel: int
+        pel: int | None
         """Parameter used by search mode for subpixel accuracy."""
 
 
@@ -422,7 +432,8 @@ class SearchMode(SearchModeBase, CustomIntEnum):
 
     @overload
     def __call__(  # type: ignore
-        self: Literal[ONETIME], step: int | tuple[int, int] = ..., pel: int = ..., /, **kwargs: Any
+        self: Literal[ONETIME], step: int | tuple[int, int] | None = ..., pel: int | None = ...,
+        recalc_mode: SearchMode = ..., /, **kwargs: Any
     ) -> SearchMode.Config:
         """
         Get the :py:class:`SearchMode.Config` from this mode and params.
@@ -435,7 +446,8 @@ class SearchMode(SearchModeBase, CustomIntEnum):
 
     @overload
     def __call__(  # type: ignore
-        self: Literal[NSTEP], times: int | tuple[int, int] = ..., pel: int = ..., /, **kwargs: Any
+        self: Literal[NSTEP], times: int | tuple[int, int] | None = ..., pel: int | None = ...,
+        recalc_mode: SearchMode = ..., /, **kwargs: Any
     ) -> SearchMode.Config:
         """
         Get the :py:class:`SearchMode.Config` from this mode and params.
@@ -448,7 +460,8 @@ class SearchMode(SearchModeBase, CustomIntEnum):
 
     @overload
     def __call__(  # type: ignore
-        self: Literal[DIAMOND], init_step: int | tuple[int, int] = ..., pel: int = ..., /, **kwargs: Any
+        self: Literal[DIAMOND], init_step: int | tuple[int, int] | None = ..., pel: int | None = ...,
+        recalc_mode: SearchMode = ..., /, **kwargs: Any
     ) -> SearchMode.Config:
         """
         Get the :py:class:`SearchMode.Config` from this mode and params.
@@ -461,7 +474,8 @@ class SearchMode(SearchModeBase, CustomIntEnum):
 
     @overload
     def __call__(  # type: ignore
-        self: Literal[HEXAGON], range: int | tuple[int, int] = ..., pel: int = ..., /, **kwargs: Any
+        self: Literal[HEXAGON], range: int | tuple[int, int] | None = ..., pel: int | None = ...,
+        recalc_mode: SearchMode = ..., /, **kwargs: Any
     ) -> SearchMode.Config:
         """
         Get the :py:class:`SearchMode.Config` from this mode and params.
@@ -474,7 +488,8 @@ class SearchMode(SearchModeBase, CustomIntEnum):
 
     @overload
     def __call__(  # type: ignore
-        self: Literal[UMH], range: int | tuple[int, int] = ..., pel: int = ..., /, **kwargs: Any
+        self: Literal[UMH], range: int | tuple[int, int] | None = ..., pel: int | None = ...,
+        recalc_mode: SearchMode = ..., /, **kwargs: Any
     ) -> SearchMode.Config:
         """
         Get the :py:class:`SearchMode.Config` from this mode and params.
@@ -488,7 +503,8 @@ class SearchMode(SearchModeBase, CustomIntEnum):
     @overload
     def __call__(  # type: ignore
         self: Literal[EXHAUSTIVE] | Literal[EXHAUSTIVE_H] | Literal[EXHAUSTIVE_V],
-        radius: int | tuple[int, int] = ..., pel: int = ..., /, **kwargs: Any
+        radius: int | tuple[int, int] | None = ..., pel: int | None = ..., recalc_mode: SearchMode = ...,
+        /, **kwargs: Any
     ) -> SearchMode.Config:
         """
         Get the :py:class:`SearchMode.Config` from this mode and params.
@@ -500,7 +516,10 @@ class SearchMode(SearchModeBase, CustomIntEnum):
         """
 
     @overload
-    def __call__(self, param: int | tuple[int, int] = ..., pel: int = ..., /, **kwargs: Any) -> SearchMode.Config:
+    def __call__(
+        self, param: int | tuple[int, int] | None = ..., pel: int | None = ..., recalc_mode: SearchMode = ...,
+        /, **kwargs: Any
+    ) -> SearchMode.Config:
         """
         Get the :py:class:`SearchMode.Config` from this mode and params.
 
@@ -511,7 +530,8 @@ class SearchMode(SearchModeBase, CustomIntEnum):
         """
 
     def __call__(
-        self, param: int | tuple[int, int] | MissingT = MISSING, pel: int | MissingT = MISSING, /, **kwargs: Any
+        self, param: int | tuple[int, int] | None | MissingT = MISSING, pel: int | None | MissingT = MISSING,
+        recalc_mode: SearchMode | MissingT = MISSING, /, **kwargs: Any
     ) -> SearchMode.Config:
         """
         Get the :py:class:`SearchMode.Config` from this mode and params.
@@ -529,25 +549,34 @@ class SearchMode(SearchModeBase, CustomIntEnum):
         if self is SearchMode.AUTO:
             self = SearchMode.DIAMOND
 
-        param_recalc: int | MissingT
+        if recalc_mode is MISSING:
+            recalc_mode = SearchMode.ONETIME
+
+        param_recalc: int | MissingT | None
 
         if isinstance(param, int):
             param, param_recalc = param, MISSING
         elif isinstance(param, tuple):
             param, param_recalc = param
         else:
-            param = param_recalc = MISSING
+            param = param_recalc = param
 
         if param is MISSING:
             param = (2 if is_uhd else 5) if (refine and truemotion) else (1 if is_uhd else 2)
 
+        param_c = fallback(param, 2)
+
         if param_recalc is MISSING:
-            param_recalc = max(0, round(exp(0.69 * param - 1.79) - 0.67))
+            param_recalc = max(0, round(exp(0.69 * param_c - 1.79) - 0.67))
 
         if pel is MISSING:
-            pel = min(8, max(0, param * 2 - 2))
+            pel = min(8, max(0, param_c * 2 - 2))
 
-        return SearchMode.Config(self, param, param_recalc, pel)  # type: ignore
+        return SearchMode.Config(self, recalc_mode, param, param_recalc, pel)  # type: ignore
+
+    @property
+    def defaults(self) -> SearchMode.Config:
+        return self(None, None, self)
 
 
 class MVTools:
@@ -870,8 +899,6 @@ class MVTools:
         vectors = MotionVectors() if inplace else self.vectors
 
         if isinstance(sad_mode, tuple):
-            if not sad_mode[1].is_satd:
-                raise CustomValueError('The SADMode for recalculation must use SATD!', self.__class__)
             sad_mode, recalc_sad_mode = sad_mode
         else:
             sad_mode, recalc_sad_mode = sad_mode, SADMode.SATD
@@ -916,20 +943,18 @@ class MVTools:
         analyze_args = dict[str, Any](
             dct=sad_mode, pelsearch=search.pel, blksize=blocksize, overlap=overlap, search=search.mode,
             truemotion=motion.truemotion, searchparam=search.param, chroma=self.chroma,
-            plevel=0, pglobal=11
+            plevel=motion.plevel, pglobal=motion.pglobal
         ) | self.analyze_args
 
         recalc_args = dict[str, Any](
-            search=0, dct=recalc_sad_mode, thsad=thSAD_recalc, blksize=halfblocksize, overlap=halfoverlap,
-            truemotion=motion.truemotion, searchparam=search.param_recalc, chroma=self.chroma
+            search=search.recalc_mode, dct=recalc_sad_mode, thsad=thSAD_recalc, blksize=halfblocksize,
+            overlap=halfoverlap, truemotion=motion.truemotion, searchparam=search.param_recalc,
+            chroma=self.chroma
         ) | self.recalculate_args
 
         if self.mvtools is MVToolsPlugin.FLOAT_NEW:
             vmulti = self.mvtools.Analyse(super_search, radius=t2, **analyze_args)
 
-            if self.source_type.is_inter:
-                vmulti = vmulti.std.SelectEvery(4, 2, 3)
-
             vectors.vmulti = vmulti
 
             for i in range(self.refine):
@@ -945,25 +970,18 @@ class MVTools:
 
                     vectors.set_mv(direction, delta, vect)
 
-            for i in range(1, self.tr + 1):
+            for i in range(1, t2 + 1):
                 _add_vector(i)
 
             if self.refine:
-                refblks = blocksize
-
                 for i in range(1, t2 + 1):
                     if not vectors.got_mv(MVDirection.BACK, i) or not vectors.got_mv(MVDirection.FWRD, i):
                         continue
 
-                    for j in range(1, self.refine):
-                        val = (refblks / 2 ** j)
+                    for j in range(0, self.refine):
+                        val = clamp(blocksize / 2 ** j, 4, 128)
 
-                        if val > 128:
-                            refblks = 128
-                        elif val < 4:
-                            refblks = blocksize
-
-                        recalc_args.update(blksize=refblks / 2 ** j, overlap=refblks / 2 ** (j + 1))
+                        recalc_args.update(blksize=val, overlap=val / 2)
 
                         _add_vector(i, False)
 
@@ -1000,6 +1018,7 @@ class MVTools:
                 vectors_forward.append(vmulti.std.SelectEvery(t2 * 2, i + 1))
         else:
             it = 1 + int(self.source_type.is_inter)
+
             for i in range(it, t2 + 1, it):
                 vectors_backward.append(vectors.get_mv(MVDirection.BACK, i))
                 vectors_forward.append(vectors.get_mv(MVDirection.FWRD, i))
@@ -1106,7 +1125,7 @@ class MVTools:
         thSAD, thSADC = (thSAD if isinstance(thSAD, tuple) else (thSAD, None))
 
         thSAD = kwargs_fallback(thSAD, (self.vectors.kwargs, 'thSAD'), 300)
-        thSADC = fallback(thSADC, round(thSAD * 0.18875 * exp(2 * 0.693)) if self.params_curve else thSAD)
+        thSADC = fallback(thSADC, round(thSAD * 0.18875 * exp(2 * 0.693)) if self.params_curve else thSAD // 2)
 
         limit, limitC = normalize_seq(limit, 2)
 
@@ -1115,6 +1134,8 @@ class MVTools:
                 '"limit" values should be between 0 and 255 (inclusive)!', self.__class__.degrain
             )
 
+        limitf, limitCf = scale_value(limit, 8, ref), scale_value(limitC, 8, ref)
+
         thSCD1, thSCD2 = thSCD
 
         thSCD1 = fallback(thSCD1, round(0.35 * thSAD + 260) if self.params_curve else thSAD // 2)
@@ -1132,9 +1153,9 @@ class MVTools:
         degrain_args = dict[str, Any](thscd1=thSCD1, thscd2=thSCD2, plane=self.mv_plane)
 
         if self.mvtools is MVToolsPlugin.INTEGER:
-            degrain_args.update(thsad=thSAD, thsadc=thSADC, limit=limit, limitc=limitC)
+            degrain_args.update(thsad=thSAD, thsadc=thSADC, limit=limitf, limitc=limitCf)
         else:
-            degrain_args.update(thsad=[thSAD, thSADC, thSADC], limit=[limit, limitC])
+            degrain_args.update(thsad=[thSAD, thSADC, thSADC], limit=[limitf, limitCf])
 
             if self.mvtools is MVToolsPlugin.FLOAT_NEW:
                 degrain_args.update(thsad2=[thSAD / 2, thSADC / 2])
